#!/usr/bin/perl

#=================================================================
# APPAL : APProximate ALigner, for use in TWAIN
#
# bmajoros@tigr.org
# 5/17/04
#=================================================================

use strict;
use Needleman;
use ConfigFile;
use FastaReader;
use FastaWriter;

my $usage="$0 <reference.fasta> <query.fasta> <*.iso> <outfile>";
die "$usage\n" unless @ARGV==4;
my ($fasta1,$fasta2,$isoFilename,$outfile)=@ARGV;
my $seq1=loadFasta($fasta1);
my $seq2=loadFasta($fasta2);
my $fastaLength1=length($$seq1);
my $fastaLength2=length($$seq2);

# Parse *.iso file to get name of *.cfg file
open(IN,$isoFilename) || die "Can't open file: $isoFilename\n";
my $configFilename;
while(<IN>){if(/<=.*:\s*(\S*\.cfg)/) {$configFilename=$1;last}}
close(IN);

##################################################################
#                          RUN NUCMER                            #
# Run nucmer on the whole sequences to get the set of all high-  #
# scoring hits                                                   #
##################################################################
my $configFile=new ConfigFile($configFilename);
my $maxGap=$configFile->lookupOrDie("nucmer-maxgap");        # default=90
my $minMatch=$configFile->lookupOrDie("nucmer-minmatch");    # default=20
my $minCluster=$configFile->lookupOrDie("nucmer-mincluster");# default=65
my $matrix=$configFile->lookupOrDie("needleman-matrix");
my $gapPenalty=$configFile->lookupOrDie("needleman-gap-penalty");
my $TWAINdir=$ENV{"TWAIN"};
die "Please 'setenv TWAIN /.../path/to/twain/.../' before running this program\n" unless defined $TWAINdir;
my $outStem="APPAL.tmp";
my $command="nucmer -l $minMatch -g $maxGap -c $minCluster -p $outStem --maxmatch $fasta1 $fasta2";
print STDERR "$command\n";
system($command);
my %deltaFile;
my $deltaFile="${outStem}.delta";
open(IN,$deltaFile) || die "Can't open $deltaFile\n";
<IN>;<IN>;<IN>;
while(<IN>)
  {
    my $header=$_;
    my @fields=split/\s+/,$_;
    die "error parsing line in $deltaFile: $_\n" unless @fields==7;
    my ($queryBegin,$queryEnd,$refBegin,$refEnd,$numIdent,
	$numSim,$numStops)=@fields;
    --$refBegin; --$queryBegin;
    my @deltas;
    while(1)
      {
	my $delta=0+<IN>;
	last unless $delta;
	push @deltas,$delta;
      }
    my $signature="$queryBegin $queryEnd $refBegin $refEnd";
    $deltaFile{$signature}={header=>$header,deltas=>\@deltas};
  }
close(IN);

##################################################################
#                        RUN SHOW-COORDS                         #
# Use "show-coords -g" to find the optimal set of nonoverlapping #
# nucmer hits                                                    #
##################################################################
$command="show-coords -r -g $deltaFile";
my (@bestHits,@gaps);
open(PIPE,"$command|") || die "Can't exec command: $command";
while(<PIPE>)
  {
    if(/\s*(\d+)\s+(\d+)\s*\|\s*(\d+)\s+(\d+)\s*\|\s*(\d+)\s+(\d+)\s*\|\s*([\d\.]+)\s*\|\s*(\S+)\s+(\S+)\s*/)
      {
	my ($S1,$E1,$S2,$E2,$len1,$len2,$pctIdent,$tag1,$tag2)=
	  ($1,$2,$3,$4,$5,$6,$7,$8,$9);
	print STDERR "show-coords selects: $S1-$E1:$S2-$E2,$pctIdent\%\n";
	my $hit=
	  {
	   start1=>$S1-1,
	   start2=>$S2-1,
	   end1=>$E1,
	   end2=>$E2,
	   length1=>$len1,
	   length2=>$len2,
	   percentIdentity=>$pctIdent,
	   tag1=>$tag1,
	   tag2=>$tag2
	  };
	push @bestHits,$hit;
      }
  }
close(PIPE);
my $numHits=@bestHits;
print STDERR "$numHits hits\n";
if($numHits)
  {
    my $firstHit=$bestHits[0];
    push @gaps,[[0,$firstHit->{start1}],[0,$firstHit->{start2}]];
    for(my $i=0 ; $i<$numHits-1 ; ++$i)
      {
	my $thisHit=$bestHits[$i];
	my $nextHit=$bestHits[$i+1];
	push @gaps,[[$thisHit->{end1},$nextHit->{start1}],
                  [$thisHit->{end2},$nextHit->{start2}]];
      }
    my $lastHit=$bestHits[$numHits-1];
    push @gaps,[[$lastHit->{end1},$fastaLength1],
	      [$lastHit->{end2},$fastaLength2]];
  }
else
  {
    push @gaps,[[0,$fastaLength1],[0,$fastaLength2]];
  }

##################################################################
#                        RUN NEEDLEMAN                           #
# Use needleman to align the regions between the optimal nucmer  #
# hits                                                           #
##################################################################
my $numGaps=@gaps;
print STDERR "$numGaps gaps\n";
my $needleman=new Needleman("$TWAINdir/alignment",$matrix);
my @gapAlignments;
for(my $i=0 ; $i<$numGaps ; ++$i)
  {
    my $gap=$gaps[$i];
    my ($interval1,$interval2)=@$gap;
    my ($begin1,$end1)=@$interval1;
    my ($begin2,$end2)=@$interval2;
    my $len1=$end1-$begin1;
    my $len2=$end2-$begin2;
    print STDERR "Running needleman on $begin1-$end1:$begin2-$end2 ($len1 vs. $len2)\n";
    my $subseq1=substr($$seq1,$begin1,$len1);
    my $subseq2=substr($$seq2,$begin2,$len2);
    my $fasta1="APPAL.tmp1.fasta";
    my $fasta2="APPAL.tmp2.fasta";
    writeFasta(\$subseq1,">1",$fasta1);
    writeFasta(\$subseq2,">2",$fasta2);

    my $alignment=$needleman->align($fasta1,$fasta2,"DNA",$gapPenalty);
    my $pctIdent=$alignment->getPercentIdentity();
    my $len=$alignment->getLength();
    my $score=$alignment->getScore();
    my $matches=$alignment->getNumMatches();
    print STDERR "$pctIdent\% identity, matches=$matches, score=$score, alignment length=$len\n";
    push @gapAlignments,$alignment;
  }


##################################################################
#                        COMBINE RESULTS                         #
# Combine results to produce a single global alignment           #
##################################################################
die unless $numGaps==$numHits+1;
my @segments; # a sequence of .appal format alignments
push @segments,@{needlemanToAppal($gapAlignments[0])};
for(my $i=1 ; $i<$numGaps ; ++$i)
  {
    push @segments,@{nucmerToAppal($bestHits[$i-1])};
    push @segments,@{needlemanToAppal($gapAlignments[$i])};
  }
coalesceAppal(\@segments);
outputAppal(\@segments);

##################################################################
#                             subs                               #
##################################################################
sub coalesceAppal
  {
    my ($appal)=@_;
    my $n=@$appal;
    my $prev=substr($appal->[0],0,1);
    for(my $i=1 ; $i<$n ; ++$i)
      {
	my $this=substr($appal->[$i],0,1);
	if($this eq $prev)
	  {
	    collapseAppal($appal,$i-1,$i);
	    --$n;
	    --$i;
	    next;
	  }
	$prev=$this;
      }
  }
sub collapseAppal
  {
    my ($appal,$prevIndex,$thisIndex)=@_;
    my $prev=$appal->[$prevIndex];
    my $this=$appal->[$thisIndex];
    $prev=~/(.)(.*)/;
    my ($prevType,$prevCount)=($1,$2);
    $this=~/(.)(.*)/;
    my ($thisType,$thisCount)=($1,$2);
    die unless $thisType eq $prevType;
    my $combinedCount=$prevCount+$thisCount;
    $appal->[$prevIndex]="$prevType$combinedCount";
    splice(@$appal,$thisIndex,1);
  }
sub outputAppal
  {
    my ($appal)=@_;
    open(OUT,">$outfile") || die "can't create file: $outfile";
    print OUT "1 $fastaLength1 1 $fastaLength2 0 0 0\n";
    my $n=@$appal;
    for(my $i=0 ; $i<$n ; ++$i)
      {
	my $elem=$appal->[$i];
	$elem=~/(.)(.*)/;
	print OUT "$1 $2\n";
      }
    print OUT "e 0\n";
    close(OUT);
  }
sub nucmerToAppal
  {
    my ($hit)=@_;
    my $start1=$hit->{start1};
    my $end1=$hit->{end1};
    my $start2=$hit->{start2};
    my $end2=$hit->{end2};
    my $len1=$hit->{length1};
    my $len2=$hit->{length2};
    my $signature="$start1 $end1 $start2 $end2";
    my $deltaRecord=$deltaFile{$signature};
    die "no delta record for $signature " unless defined $deltaRecord;
    my $deltas=$deltaRecord->{deltas};
    my $numDeltas=0+@$deltas;
    my (@appal,$x,$y);
    for(my $i=0 ; $i<$numDeltas ; ++$i)
      {
	my $delta=$deltas->[$i];
	my $sign=($delta>0 ? "+" : "-");
	$delta=abs($delta);
	--$delta;
	if($delta>0) {push @appal,"m$delta";$x+=$delta;$y+=$delta}
	if($sign eq "+") {push @appal,"i1";++$x}
	else {push @appal,"d1";++$y}
      }
    my $dx=$len1-$x;#+1;
    my $dy=$len2-$y;#+1;
    die "$x $len1 $dx , $y $len2 $dy " unless $dx==$dy;
    push @appal,"m$dx";
    return \@appal;
  }
sub needlemanToAppal
  {
    my ($alignment)=@_;
    my $aln=$alignment->getAlignment();
    my $aln0=$aln->[0];
    my $aln1=$aln->[1];
    my $aln2=$aln->[2];
    my $len=length $aln0;
    my @appal;
    for(my $i=0 ; $i<$len ; ++$i)
      {
	if(substr($aln0,$i,1) eq '-') {push @appal,"d1"}
	elsif(substr($aln2,$i,1) eq '-') {push @appal,"i1"}
	else {push @appal,"m1"}
      }
    return \@appal;
  }
sub getLength
  {
    my ($filename)=@_;
    return FastaReader::getGenomicSize($filename);
  }
sub loadFasta
  {
    my ($filename)=@_;
    my $reader=new FastaReader($filename);
    my ($def,$seq)=$reader->nextSequence();
    return \$seq;
  }
sub writeFasta
  {
    my ($seq,$def,$filename)=@_;
    open(OUT,">$filename") || die "Can't write to $filename";
    print OUT "$def\n$$seq\n";
    close(OUT);
  }


