/****************************************************************
 OasisMatrix.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "OasisMatrix.H"
#include <iostream>
#include "tigr++/TigrArray1DSorter.H"



/****************************************************************
                        CellLink methods
 ****************************************************************/
CellLink::CellLink()
  : packedByte(static_cast<unsigned char>(0))
{
}



CellLink::CellLink(int queryIndex,int referenceIndex,int queryPhase,
		   int referencePhase,float score)
  : queryIndex(queryIndex),
    referenceIndex(referenceIndex),
    inductiveScore(score)
{
  setQueryPhase(queryPhase);
  setReferencePhase(referencePhase);
  setActive();
}



bool CellLink::isActive()
{
  return packedByte & ACTIVE_MASK;
}



void CellLink::setActive()
{
  packedByte|=ACTIVE_MASK;
}



int CellLink::getReferencePhase() const
{
  return packedByte & QUERY_PHASE_MASK;
}



int CellLink::getQueryPhase() const
{
  return (packedByte & REF_PHASE_MASK) >> 2;
}



void CellLink::setReferencePhase(int f)
{
  packedByte=(packedByte & QUERY_PHASE_NEG) | f;
}



void CellLink::setQueryPhase(int f)
{
  packedByte=(packedByte & REF_PHASE_NEG) | (f<<2);
}



int CellLink::getLinkIndex(int referencePhase,int queryPhase)
{
  return referencePhase+3*queryPhase;
}




/****************************************************************
                           Cell methods
 ****************************************************************/
Cell::Cell(Signal &referenceSignal,Signal &querySignal,int numPromerHits)
  : referenceSignal(referenceSignal),
    querySignal(querySignal),
    queryCount(0),
    promerHits(numPromerHits)
{
}



int Cell::getReferenceIndex()
{
  return referenceSignal.getIndex();
}



int Cell::getQueryIndex()
{
  return querySignal.getIndex();
}



TigrArray1D<CellLink> &Cell::allocateLinks(int howMany)
{
  links=new TigrArray1D<CellLink>(howMany);
  return *links;
}



bool Cell::referenceAtLeftTerminus()
{
  return referenceSignal.isLeftTerminus();
}



bool Cell::queryAtLeftTerminus()
{
  return querySignal.isLeftTerminus();
}



CellLink *Cell::getLink(int queryPhase,int referencePhase)
{
  // First, construct a compact search image, to avoid having
  // to do bit operations over and over again
  unsigned char rPhase=static_cast<unsigned char>(queryPhase);
  unsigned char qPhase=static_cast<unsigned char>(referencePhase);
  unsigned char searchImage=CellLink::ACTIVE_MASK | qPhase | (rPhase<<2);
  
  // Now find this link using a binary search
  TigrArray1D<CellLink> &array=*links;
  int min=0, max=array.size()-1;
  while(min<max)
    {
      int mid=(min+max)/2;
      if(searchImage>array[mid].getLinkSignature()) min=mid+1;
      else max=mid;
    }
  if(min>=array.size()) return NULL;
  CellLink &link=array[min];
  return link.getLinkSignature()==searchImage ? &link : NULL;
}



void Cell::decrementQueryCount()
{
  --queryCount;
  if(queryCount<1)
    {
      int n=links->size();
      CellLink *pLink=&(*links)[0];
      for(int i=0 ; i<n ; ++i)
	{
	  CellLink &link=*pLink;
	  ++pLink;
	}
    }
}



bool Cell::leftTermini() 
{
  return referenceAtLeftTerminus()&&queryAtLeftTerminus();
}



void Cell::printOn(ostream &os) const
{
  os<<referenceSignal.getSignalType()<<"@"<<referenceSignal.getConsensusPosition()
    <<" x "
    <<querySignal.getSignalType()<<"@"<<querySignal.getConsensusPosition();
}



ostream &operator<<(ostream &os,const Cell &cell)
{
  cell.printOn(os);
  return os;
}



/****************************************************************
                          RowCol methods
 ****************************************************************/
RowCol::RowCol(int size)
  : cells(size)
{
}



Cell *&RowCol::operator[](int i)
{
  return cells[i];
}



int RowCol::size()
{
  return cells.size();
}



void RowCol::sort()
{
  ColumnCellComparator cmp;
  TigrArray1DSorter<Cell*> sorter(cells,cmp);
  sorter.sortAscendInPlace();
}



bool RowCol::ColumnCellComparator::equal(Cell *&a,Cell *&b)
{
  return a->referenceSignal.getIndex() == b->referenceSignal.getIndex();
}



bool RowCol::ColumnCellComparator::greater(Cell *&a,Cell *&b)
{
  return a->referenceSignal.getIndex() > b->referenceSignal.getIndex();
}



bool RowCol::ColumnCellComparator::less(Cell *&a,Cell *&b)
{
  return a->referenceSignal.getIndex() < b->referenceSignal.getIndex();
}



bool RowCol::singleton() 
{
  return size()==1;
}



/****************************************************************
                         Column methods
 ****************************************************************/
Cell *Column::lookupReferenceIndex(int rowID)
{
  int min=0, max=size();
  if(max==0) return NULL;
  while(max-min>1)
    {
      int mid=(min+max)/2;
      Cell *cell=cells[mid];
      if(rowID>=cell->getReferenceIndex()) min=mid;
      else max=mid;
    }
  Cell *cell=cells[min];
  return cell->getReferenceIndex()==rowID ? cell : NULL;
}



/****************************************************************
                       OasisMatrix methods
 ****************************************************************/
OasisMatrix::OasisMatrix(ParseGraph &referenceGraph,ParseGraph &queryGraph)
  : referenceGraph(referenceGraph),
    queryGraph(queryGraph),
    columns(queryGraph.numVertices()),
    numCells(0)
{
  columns.setAllTo(NULL);
}



Cell *OasisMatrix::getCell(int queryIndex,int referenceIndex)
{
  Column &column=*columns[queryIndex];
  int min=0, max=column.size();
  while(min+1<max)
    {
      int mid=(min+max)/2;
      Cell *cell=column[mid];
      int midVal=cell->getReferenceIndex();
      if(referenceIndex<midVal) max=mid;
      else min=mid;
    }

  Cell *cell=column[min];
  return cell->getReferenceIndex()==referenceIndex ? cell : NULL;
}



Column &OasisMatrix::allocateColumn(int index,int size)
{
  numCells+=size;
  return *(columns[index]=new Column(size));
}



int OasisMatrix::numReferenceSignals()
{
  return referenceGraph.numVertices();
}



int OasisMatrix::numQuerySignals()
{
  return queryGraph.numVertices();
}



Column *OasisMatrix::getColumn(int queryIndex)
{
  return columns[queryIndex];
}



int OasisMatrix::getNumCells()
{
  return numCells;
}



bool OasisMatrix::isOnPromerHit(Cell &c) 
{
  return columns[c.getQueryIndex()]->singleton();
}


