/****************************************************************
 OasisMatrix.H : the 2D dynamic programming matrix for Oasis
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#ifndef INCL_OasisMatrix_H
#define INCL_OasisMatrix_H
#include "tigr++/TigrArray1D.H"
#include "tigr++/TigrComparator.H"
#include "tigr++/TigrBitSet.H"
#include "tigrscan/ParseGraph.H"
#include <iostream>
using namespace std;


/****************************************************************
                         class CellLink
 A link from one cell in the DP matrix back to a previous cell.
 These links make up the 2D "trellis" (not graph) which allows
 identification of the optimal path through the matrix.
 ****************************************************************/
class CellLink
{
  unsigned char packedByte; 
     // ={isActive:1bit,queryPhase:2bits,referencePhase:2bits}
     // [note that phases refer to the rhs end of the link]

  short unsigned int referenceIndex, queryIndex; // address of pred cell
  float inductiveScore;

  static const unsigned char QUERY_PHASE_MASK=3;  // 00000011
  static const unsigned char QUERY_PHASE_NEG=0xFC;// 11111100
  static const unsigned char REF_PHASE_MASK=12;   // 00001100
  static const unsigned char REF_PHASE_NEG=0xF3;  // 11110011
  static const unsigned char PHASES_MASK=15;      // 00001111
public:
  static const unsigned char ACTIVE_MASK=16;      // 00010000
  CellLink();
  CellLink(int queryIndex,int referenceIndex,int queryPhase,
	   int referencePhase,float score);
  inline short unsigned int getReferenceIndex() const   
    {return referenceIndex;}
  inline short unsigned int getQueryIndex() const {return queryIndex;}
  inline float getInductiveScore() const          {return inductiveScore;}
  inline void setReferenceIndex(short unsigned int i){referenceIndex=i;}
  inline void setQueryIndex(short unsigned int i)    {queryIndex=i;}
  inline void setInductiveScore(float s)             {inductiveScore=s;}
  inline int getLinkIndex() {return getReferencePhase()+3*getQueryPhase();}
  static int getLinkIndex(int referencePhase,int queryPhase);
  inline unsigned char getLinkSignature() {return packedByte;}
  bool isActive(); // default=inactive
  void setActive();
  int getReferencePhase() const;
  int getQueryPhase() const;
  void setReferencePhase(int);
  void setQueryPhase(int);
};



/****************************************************************
                           class Cell
          A single cell in the dynamic programming matrix.
 ****************************************************************/
struct Cell
{
  TigrArray1D<CellLink> *links;
  Signal &referenceSignal, &querySignal;
  int queryCount;
  TigrBitSet promerHits;

  Cell(Signal &referenceSignal,Signal &querySignal,int numPromerHits);
  int getReferenceIndex();// index into parse graph's array of vertices
  int getQueryIndex();// index into parse graph's array of vertices
  TigrArray1D<CellLink> &allocateLinks(int howMany);
  bool referenceAtLeftTerminus();
  bool queryAtLeftTerminus();
  bool leftTermini();
  CellLink *getLink(int queryPhase,int referencePhase);
  void decrementQueryCount();
  void printOn(ostream &) const;
};
ostream &operator<<(ostream &,const Cell &);



/****************************************************************
                        class Row/Column
 Originally was going to represent either rows or columns in the
 matrix, but I only use it for columns.  Thus, you can isolate an
 entire column, but not (explicitly) an entire row.
 ****************************************************************/
class RowCol
{
protected:
  TigrArray1D<Cell*> cells;
  struct ColumnCellComparator : public TigrComparator<Cell*>
  {
    virtual bool equal(Cell* &a,Cell* &b);
    virtual bool greater(Cell* &a,Cell* &b);
    virtual bool less(Cell* &a,Cell* &b);
  };
public:
  RowCol(int size);
  Cell *&operator[](int);
  int size();
  void sort();
  bool singleton();
};



// ### Should be folded into RowCol to replace the latter...
struct Column : public RowCol 
{ 
  Column(int size) : RowCol(size) {} 
  Cell *lookupReferenceIndex(int);
};



/****************************************************************
                         class OasisMatrix

  Represents a sparse dynamic programming matrix.  Consists of 
  Column objects (which consist of Cells).

  Remember that rows (y) are associated with the reference genome, and
  columns (x) are associated with the query genome.
 ****************************************************************/
class OasisMatrix
{
  TigrArray1D<Column*> columns; // indexed by signals in query genome
  ParseGraph &referenceGraph, &queryGraph;
  int numCells;
public:
  OasisMatrix(ParseGraph &referenceGraph,ParseGraph &queryGraph);
  Cell *getCell(int queryIndex,int referenceIndex);
  Column &allocateColumn(int index,int size);
  int numReferenceSignals();
  int numQuerySignals();
  inline int numRows()              {return numReferenceSignals();}
  inline int numColumns()             {return numQuerySignals();}
  Column *getColumn(int queryIndex);
  int getNumCells();
  bool isOnPromerHit(Cell &);
};


#endif
