/****************************************************************
 PromerParser.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "PromerParser.H"
#include <iostream>
#include "tigr++/TigrFile.H"
#include "tigr++/TigrVectorSorter.H"


/**************************************************************
                      class PromerParser
***************************************************************/
PromerParser::PromerParser(int tolerance,int minPercentIdentity,
			   int minPercentSimilarity,int maxPercentStop)
  : tolerance(tolerance),
    minIdent(minPercentIdentity),
    minSim(minPercentSimilarity),
    maxStop(maxPercentStop)
{
}



PromerCoordMap *PromerParser::parseDelta(const TigrString &filename)
{
  PromerCoordMap *coordMap=new PromerCoordMap(tolerance);
  TigrFile file(filename);
  file.getline(); // just lists the two FASTA files
  file.getline(); // "PROMER"
  TigrString seqHeader=file.getline(); // sequence ID's
  if(file.eof()) return NULL;
  seqHeader.trimWhitespace();
  if(seqHeader.isEmpty()) return NULL;
  while(!file.eof())
    {
      // Parse match header
      TigrString line=file.getline();
      if(file.eof()) break;
      TigrVector<TigrString> &fields=*line.getFields();
      if(fields.size()<7) {throw line;}
      int referenceBegin=fields[0].asInt();
      int referenceEnd=fields[1].asInt();
      int queryBegin=fields[2].asInt();
      int queryEnd=fields[3].asInt();
      int numIdent=fields[4].asInt();
      int numSim=fields[5].asInt();
      int numStops=fields[6].asInt();
      Strand referenceStrand, queryStrand;
      if(referenceBegin<referenceEnd)
	  referenceStrand=FORWARD_STRAND;
      else
	{
	  referenceStrand=REVERSE_STRAND;
	  swap(referenceBegin,referenceEnd);
	}
      if(queryBegin<queryEnd)
	queryStrand=FORWARD_STRAND;
      else
	{
	  queryStrand=REVERSE_STRAND;
	  swap(queryBegin,queryEnd);
	}
      --referenceBegin; --queryBegin;//convert to 0/S-based  coords
      int nonIdentities=fields[4].asInt();
      int nonSimilarities=fields[5].asInt();
      int stopCodons=fields[6].asInt();
      delete &fields;
      PromerMatch *match=
	new PromerMatch(queryBegin,queryEnd,referenceBegin,referenceEnd,
			queryStrand,referenceStrand);
      short int queryMultiplier=match->getQueryMultiplier();
      short int referenceMultiplier=match->getReferenceMultiplier();

      // Read distances between sequential indels
      int referencePos=referenceStrand==
	FORWARD_STRAND ? referenceBegin : referenceEnd;
      int queryPos=queryStrand==FORWARD_STRAND ? queryBegin : queryEnd;
      match->addNode(new PromerDeltaNode(queryPos,referencePos));
      int alignmentLength=referenceEnd-referenceBegin;
      while(true)
	{
	  int delta=file.getline().asInt()*3;
	  if(delta==0) break;
	  int absDelta=abs(delta);
	  referencePos+=absDelta*referenceMultiplier;
	  queryPos+=absDelta*queryMultiplier;
	  
	  if(delta>0) 
	    {
	      queryPos-=3*queryMultiplier;
	      alignmentLength+=3;
	    }
	  else 
	    referencePos-=3*referenceMultiplier;

	  PromerDeltaNode *newNode=
	    new PromerDeltaNode(queryPos,referencePos);
	  if(delta==-3) match->discardPrevious();
	  match->addNode(newNode);
	}
      short int pctIdent=100-
	static_cast<short int>(300*numIdent/float(alignmentLength)+0.5);
      short int pctSim=100-
	static_cast<short int>(300*numSim/float(alignmentLength)+0.5);
      short int pctStops=
	static_cast<short int>(300*numStops/float(alignmentLength)+0.5);
      if(pctIdent<minIdent || pctSim<minSim || pctStops>maxStop)
	delete match;
      else
	{
	  match->setPercents(pctIdent,pctSim,pctStops);
	  coordMap->addMatch(match);
	}
    }
  coordMap->doneAddingMatches();
  return coordMap;
}



/**************************************************************
                     class PromerMatch
***************************************************************/
PromerMatch::PromerMatch(int queryBegin,int queryEnd,int qBegin,int qEnd,
			 Strand queryStrand,Strand referenceStrand)
  : queryBegin(queryBegin), 
    queryEnd(queryEnd),
    referenceBegin(qBegin),
    referenceEnd(qEnd),
    referenceStrand(referenceStrand),
    queryStrand(queryStrand),
    queryMultiplier(queryStrand==FORWARD_STRAND ? 1 : -1),
    referenceMultiplier(referenceStrand==FORWARD_STRAND ? 1 : -1)
{
}



void PromerMatch::setPercents(short int percentIdent,short int percentSim, 
			      short int percentStop)
{
  this->percentIdent=percentIdent;
  this->percentSim=percentSim;
  this->percentStop=percentStop;
}



short int PromerMatch::getPercentIdent() const
{
  return percentIdent;
}



short int PromerMatch::getPercentSim() const
{
  return percentSim;
}



short int PromerMatch::getPercentStops() const
{
  return percentStop;
}



short int PromerMatch::getQueryMultiplier() const
{
  return queryMultiplier;
}



short int PromerMatch::getReferenceMultiplier() const
{
  return referenceMultiplier;
}



Strand PromerMatch::getQueryStrand() const
{
  return queryStrand;
}



Strand PromerMatch::getReferenceStrand() const
{
  return referenceStrand;
}



PromerMatch::~PromerMatch()
{
  TigrVector<PromerDeltaNode*>::iterator cur=deltaNodes.begin(), 
    end=deltaNodes.end();
  for(; cur!=end ; ++cur)
    delete *cur;
}



void PromerMatch::discardPrevious()
{
  int n=deltaNodes.size();
  if(n>0) deltaNodes.resize(n-1);
}



void PromerMatch::addNode(PromerDeltaNode *node)
{
  deltaNodes.push_back(node);
}


/******************************************************************
  This method maps a query coordinate to a reference coordinate.
  It does this by searching through the match nodes (which anchor
  the alignment at matched residues) to find the nearest match node
  to the left of this coordinate.  If the query coordinate falls
  exactly on one of these match nodes, the mapping is obviously 
  a simple one -- just map it to the reference coordinate stored in
  the match node.  Otherwise we use a linear extrapolation from the
  anchor node, using a slope of 1.  In this way, we can map a query
  coordinate even if it falls outside the query domain -- in the
  latter case we simply have a negative delta which still gives a
  reasonable (inexact) mapping to a reference coordinate.
 */
int PromerMatch::map(int queryCoord)
{
  switch(queryStrand)
    {
    case FORWARD_STRAND:
      {
	// Perform a binary search on this forward-sorted list to find
	// that element that the queryCoord is just greater than
	int min=0, max=deltaNodes.size();
	while(max-min>1)
	  {
	    int mid=(min+max)/2;
	    PromerDeltaNode &midNode=*deltaNodes[mid];
	    int midQuery=midNode.getQueryCoord();
	    if(queryCoord>=midQuery) min=mid;
	    else max=mid;
	  }
	PromerDeltaNode &node=*deltaNodes[min];
	int delta=queryCoord-node.getQueryCoord();
	return node.getReferenceCoord()+delta*referenceMultiplier;
      }

    case REVERSE_STRAND:
      {
	// Perform a binary search on this reverse-sorted list to find
	// that element that the queryCoord is just less than
	int max=0, min=deltaNodes.size();
	while(min-max>1)
	  {
	    int mid=(min+max)/2;
	    PromerDeltaNode &midNode=*deltaNodes[mid];
	    int midQuery=midNode.getQueryCoord();
	    if(queryCoord<=midQuery) max=mid;
	    else min=mid;
	  }
	PromerDeltaNode &node=*deltaNodes[max];
	int delta=node.getQueryCoord()-queryCoord;
	return node.getReferenceCoord()+delta*referenceMultiplier;
      }
    }
}



int PromerMatch::getQueryBegin() const
{
  return queryBegin;
}



int PromerMatch::getQueryEnd() const
{
  return queryEnd;
}



int PromerMatch::getReferenceBegin() const
{
  return referenceBegin;
}



int PromerMatch::getReferenceEnd() const
{
  return referenceEnd;
}



/**************************************************************
                     class PromerCoordMap
***************************************************************/
PromerCoordMap::PromerCoordMap(int tolerance)
  : tolerance(tolerance)
{
}



PromerCoordMap::~PromerCoordMap()
{
  TigrVector<PromerMatch*>::iterator cur=matches.begin(), end=matches.end();
  for(; cur!=end ; ++cur)
    delete *cur;
}



void PromerCoordMap::addMatch(PromerMatch *m)
{
  matches.push_back(m);
}



void PromerCoordMap::doneAddingMatches()
{
  // Sort the matches by begin coordinate, to help speed up the
  // search a little bit

  MatchComparator cmp;
  TigrVectorSorter<PromerMatch*> sorter(matches,cmp);
  sorter.sortAscendInPlace();
}



TigrVector<PromerMatch*> *PromerCoordMap::lookup(int queryCoord)
{
  TigrVector<PromerMatch*> *v=new TigrVector<PromerMatch*>;
  lookup(queryCoord,*v);
  return v;
}



void PromerCoordMap::lookup(int queryCoord,TigrVector<PromerMatch*> &retval)
{
  TigrVector<PromerMatch*>::iterator cur=matches.begin(), end=matches.end();
  for(; cur!=end ; ++cur)
    {
      PromerMatch *match=*cur;
      if(match->getQueryBegin()-tolerance>queryCoord) return;
      if(match->getQueryEnd()+tolerance>queryCoord) 
	retval.push_back(match);
    }
}



void PromerCoordMap::getNearestMatches(int queryCoord,PromerMatch *&left,
				       PromerMatch *&right)
{
  left=right=NULL;
  int bestLeftDistance, bestRightDistance;
  TigrVector<PromerMatch*>::iterator cur=matches.begin(), end=matches.end();
  for(; cur!=end ; ++cur)
    {
      PromerMatch *match=*cur;
      int leftDistance=queryCoord-match->getQueryEnd();
      int rightDistance=match->getQueryBegin()-queryCoord;
      if(leftDistance>0 && 
	 (left==NULL || leftDistance<bestLeftDistance))
	{
	  left=match;
	  bestLeftDistance=leftDistance;
	}
      else if(rightDistance>0 && 
	      (right==NULL || rightDistance<bestRightDistance))
	{
	  right=match;
	  bestRightDistance=rightDistance;
	}
    }  
}



/**************************************************************
                    class PromerDeltaNode
***************************************************************/
PromerDeltaNode::PromerDeltaNode(int queryCoord,int referenceCoord)
  : queryCoord(queryCoord),
    referenceCoord(referenceCoord)
{
}



int PromerDeltaNode::getQueryCoord() const
{
  return queryCoord;
}



int PromerDeltaNode::getReferenceCoord() const
{
  return referenceCoord;
}


