/****************************************************************
 PromerParser.H : parses output of the PROmer program
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#ifndef INCL_PromerParser_H
#define INCL_PromerParser_H
#include "tigr++/TigrVector.H"
#include "tigr++/TigrString.H"
#include "tigr++/TigrComparator.H"
#include "tigrscan/Strand.H"

/*********************************************************************
This class (PromerParser) can be used to parse the delta file produced 
by the PROmer program.  NOTE however, that it discards any promer match 
between opposite strands (i.e., forward on one genome vs. reverse on the 
other).  This is suitable for the OASIS syntenic genefinder, but is likely
unsuitable for other uses.  Modifications to this class to allow
these cross-strand matches should be entirely trivial.

The result of parsing the delta file is an object that maps coordinates
in the query genome to coordinates in the reference genome.
**********************************************************************/


/**************************************************************
                    class PromerDeltaNode
***************************************************************/
class PromerDeltaNode
{
  int queryCoord;
  int referenceCoord;
public:
  PromerDeltaNode(int queryCoord,int referenceCoord);
  inline int getQueryCoord() const;
  inline int getReferenceCoord() const;
};



/**************************************************************
                     class PromerMatch
***************************************************************/
class PromerMatch
{
  TigrVector<PromerDeltaNode*> deltaNodes;
  int queryBegin, queryEnd;
  int referenceBegin, referenceEnd;
  Strand queryStrand, referenceStrand;
  short int queryMultiplier, referenceMultiplier; //+1 or -1 based on strand
  short int percentIdent, percentSim, percentStop;
public:
  PromerMatch(int queryBegin,int queryEnd,int referenceBegin,
	      int referenceEnd,Strand queryStrand,Strand referenceStrand);
  virtual ~PromerMatch();
  void setPercents(short int percentIdent,short int percentSim,
		   short int percentStop);
  bool isParallel() { return referenceStrand==queryStrand; }
  void addNode(PromerDeltaNode *);
  void discardPrevious();
  int map(int queryCoord);
  int getQueryBegin() const;
  int getQueryEnd() const;
  int getQueryLength() const {return queryEnd-queryBegin;}
  int getReferenceBegin() const;
  int getReferenceEnd() const;
  Strand getQueryStrand() const;
  Strand getReferenceStrand() const;
  short int getQueryMultiplier() const;
  short int getReferenceMultiplier() const;
  short int getPercentIdent() const;
  short int getPercentSim() const;
  short int getPercentStops() const;
};



class MatchComparator : public TigrComparator<PromerMatch*>
{
public:
  virtual bool equal(PromerMatch* &a,PromerMatch* &b)
    {return a->getQueryBegin()==b->getQueryBegin();}
  virtual bool greater(PromerMatch* &a,PromerMatch* &b)
    {return a->getQueryBegin()>b->getQueryBegin();}
  virtual bool less(PromerMatch* &a,PromerMatch* &b)
    {return a->getQueryBegin()<b->getQueryBegin();}
};



/**************************************************************
                     class PromerCoordMap
***************************************************************/
class PromerCoordMap
{
  TigrVector<PromerMatch*> matches;
  int tolerance;
public:
  PromerCoordMap(int tolerance);
  virtual ~PromerCoordMap();
  void addMatch(PromerMatch *);
  void doneAddingMatches();
  TigrVector<PromerMatch*> *lookup(int queryCoord);
  void lookup(int queryCoord,TigrVector<PromerMatch*> &);
  void getNearestMatches(int queryCoord,PromerMatch *&left,
			 PromerMatch *&right);
  int numMatches() {return matches.size();}
};



/**************************************************************
                      class PromerParser
***************************************************************/
class PromerParser
{
  int tolerance;
  int minIdent, minSim, maxStop;
public:
  PromerParser(int matchEdgeTolerance,int minPercentIdentity,
	       int minPercentSimilarity,int maxPercentStop);
  PromerCoordMap *parseDelta(const TigrString &filename);
};


#endif
