/****************************************************************
 PromerSignalMap.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "PromerSignalMap.H"
#include <iostream>


PromerSignalMap::PromerSignalMap(ParseGraph &referenceSequenceGraph,
				 PromerCoordMap &M,
				 int exactMatchTolerance,
				 int nonexactMatchTolerance,
				 bool constrainNonexactMatches)
  : qGraph(referenceSequenceGraph),
    M(M),
    exactMatchTolerance(exactMatchTolerance),
    nonexactMatchTolerance(nonexactMatchTolerance),
    numVertices(referenceSequenceGraph.numVertices()),
    constrainNonexactMatches(constrainNonexactMatches)
{
}



TigrSet<SignalPtr> *PromerSignalMap::mapSignal(Signal &querySignal,
					     bool &withinMatch)
{
  TigrSet<SignalPtr> *signals=new TigrSet<SignalPtr>;
  withinMatch=mapSignal(querySignal,*signals);
  return signals;
}



bool PromerSignalMap::mapSignal(Signal &querySignal,
				TigrSet<SignalPtr> &
				  referenceSequenceSignals)
{
  int queryPos=querySignal.getConsensusPosition();
  TigrVector<PromerMatch*> matches;
  M.lookup(queryPos,matches);
  if(matches.size()==0) return false;
  TigrSet<int> referencePositions;
  TigrVector<PromerMatch*>::iterator cur=matches.begin(), end=matches.end();
  for(; cur!=end ; ++cur)
    {
      PromerMatch &m=**cur;
      int qPos=m.map(queryPos);
      referencePositions.insert(qPos);
    }

  TigrSet<int>::iterator sCur=referencePositions.begin(), 
    sEnd=referencePositions.end();
  SignalType signalType=querySignal.getSignalType();
  for(; sCur!=sEnd ; ++sCur)
    {
      int qPos=*sCur;
      int numFound=
	findSignals(qPos-exactMatchTolerance,qPos+exactMatchTolerance+1,
		    referenceSequenceSignals,signalType);
      if(numFound==0 && constrainNonexactMatches)
	numFound=
	  findSignals(qPos-nonexactMatchTolerance,
		      qPos+nonexactMatchTolerance+1,
		      referenceSequenceSignals,NO_SIGNAL_TYPE);
    }
  return true;
}



int PromerSignalMap::findSignals(int begin,int end,
				 TigrSet<SignalPtr> &signals,
				  SignalType signalType)
{
  // First, use a binary search to find the leftmost signal >= begin
  int min=0, max=numVertices;
  while(min<max)
    {
      int mid=(min+max)/2;
      Signal &s=*qGraph.getIthVertex(mid);
      int midPos=s.getConsensusPosition();
      if(begin>midPos) min=mid+1;
      else max=mid;
    }
  if(min==numVertices) return 0;

  // Now add all those after this that fall within the search region
  Signal *s=qGraph.getIthVertex(min);
  int numInserted=0;
  while(min<numVertices && s->getConsensusPosition()<end)
    {
      if(s->getSignalType()==signalType || signalType==NO_SIGNAL_TYPE)
	{
	  if(!signals.isMember(s)) ++numInserted;
	  signals.insert(s);
	}
      ++min;
      s=qGraph.getIthVertex(min);
    }
  return numInserted;
}



int PromerSignalMap::rowNearReferenceCoord(int referenceCoord)
{
  int min=0, max=numVertices;
  while(min<max)
    {
      int mid=(min+max)/2;
      Signal &s=*qGraph.getIthVertex(mid);
      int midPos=s.getConsensusPosition();
      if(referenceCoord>midPos) min=mid+1;
      else max=mid;
    }
  return min;
}


