/****************************************************************
 study-approx-align.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

/*

 RESULTS:   For %identity>=50% (approx. align score), correlation between
            approx. align score and needleman-wunsch score is 92.39%
            (but note that the slope is 0.8398, not 1.0).  Only ORFs were
            used, and only if their lengths did not differ by more than 5%.
            Alignment was over DNA (even though they were ORFs, I could not
	    use protein alignments because I could not count on having
	    promer hits for all pairs of ORFs).
*/


#include <math.h>
#include <iostream>
#include <fstream>
#include <typeinfo>
#include "tigr++/TigrExceptions.H"
#include "tigr++/TigrRandom.H"
#include "oasis.H"
#include "alignment/ApproximateAlignment.H"
#include "alignment/Needleman.H"
#include "tigr++/DiscreteLogDistr.H"

/****************************************************************
 Oasis uses an explicit graph representation of each genome's
 open reading frames (but just the high-scoring ones).
 */
#ifndef EXPLICIT_GRAPHS
#error Please edit the file tigrscan.H, uncomment the definition of EXPLICIT_GRAPHS, issue a "make clean", and recompile this project
#endif


/****************************************************************
 Report memory usage statistics.  Only works for Linux...
 */
#ifdef REPORT_MEMORY_USAGE
#include "tigr++/MemoryProfiler.H"
#endif


/****************************************************************
 Allow inserted/deleted introns (ie., nonconserved intron/exon
 structure.  Currently not supported, so leave this commented out.
 */
//#define ALLOW_INSERTIONS


/****************************************************************
 Some global variables
*/
static const char *PROGRAM_NAME="study-approx-align";
static const char *VERSION="1.0";
Alphabet alphabet;
int frame; // ### CAUTION: this is required by older code; to be removed
int gapPenalty;

void collectEdges(ParseGraph &,TigrVector<Edge*> &);
double approxAlign(int xBegin,int xEnd,int yBegin,int yEnd,
		   ApproximateAlignment<FLOAT_OR_DOUBLE> &);
double needleAlign(int xBegin,int xEnd,int yBegin,int yEnd,
		   Sequence &s1,Sequence &s2,
		   const SubstitutionMatrix<FLOAT_OR_DOUBLE> &M,
		   FLOAT_OR_DOUBLE gapExtendPenalty);


/**************************************************************************
 main() -- just handles exceptions
 */
int main(int argc,char *argv[])
  {
    try
      {
	Oasis app;
	return app.main(argc,argv);
      }
    catch(const char *p)
      {
	cerr << p << endl;
      }
    catch(const string &msg)
      {
	cerr << msg.c_str() << endl;
      }
    catch(const exception &e)
      {
	cerr << "STL exception caught in main:\n" << e.what() << endl;
      }
    catch(const RootException &e)
      {
	cerr << e.getMessage() << endl;
      }
    catch(...)
      {
	cerr << "Unknown exception caught in main" << endl;
      }
    return -1;
  }



/**************************************************************************
 Constructor for Oasis class
 */
Oasis::Oasis()
  : isoRegex("([^/]+).iso")
  {
  }



/**************************************************************************
 Oasis::main() -- implements top-level algorithm
 */
int Oasis::main(int argc,char *argv[])
  {
    //---------------------------------------------------------------------
    // Process command line
    //---------------------------------------------------------------------
    TigrCommandLine cmd(argc,argv,"aq:r:m:"); // a = ab initio predictions
    if(cmd.numArgs()!=10)
      throw string(
"study-approx-align <reference.fasta> <query.fasta> <reference.iso> <query.iso> \n\
       <*.delta> <nuc.appal> <amino.appal> <r-out.gff> <q-out.gff> <NUM-POINTS>\n\
where:\n\
      <reference.fasta> & <query.fasta> are syntenic DNA sequences, \n\
            in same orientation\n\
      <*.iso> is the isochore definition file (points to *.cfg files)\n\
      <*.delta> is the promer output file\n\
      <r-out.gff> & <q-out.gff> are names of output files\n\
      -r <*.gff> loads reference gff file (DEBUGGING)\n\
      -q <*.gff> loads query gff file (DEBUGGING)\n\
      -m <outfile> prints sparse DP matrix into outfile (then use draw-graph.pl)\n\
");
    TigrString referenceFile=cmd.arg(0);
    TigrString queryFile=cmd.arg(1);
    TigrString referenceIso=cmd.arg(2);
    TigrString queryIso=cmd.arg(3);
    TigrString deltaFile=cmd.arg(4);
    TigrString nucAppalFile=cmd.arg(5);
    TigrString aminoAppalFile=cmd.arg(6);
    TigrString rOutFile=cmd.arg(7);
    TigrString qOutFile=cmd.arg(8);
    abInitio=cmd.option('a');
    if(!isoRegex.search(referenceIso)) throw "ISO file must end in .iso";
    referenceSpecies=isoRegex[1];
    if(!isoRegex.search(queryIso)) throw "ISO file must end in .iso";
    querySpecies=isoRegex[1];
    if(cmd.option('m')) matrixOutFile=cmd.optParm('m');
    int numPoints=cmd.arg(9).asInt();
    randomize();

    //----------------------------------------------------------------------
    // First, parse the two sequences independently into graphs representing
    // all highly-scoring open reading frames.  Note that the makefile defines
    // FORWARD_STRAND_ONLY when compiling TigrScan.C, so reverse-strand signals
    // and edges are never allocated within OASIS.
    //----------------------------------------------------------------------

    float referenceGC, queryGC;
    int transcriptId=1;
    TigrScan tigrScan1("oasis","1.0",ef,transcriptId);
    referenceGraph=&tigrScan1.parse(referenceFile,referenceIso,referenceSeq,
				referenceSeqStr,referenceGC);
    stopCodons=&tigrScan1.getSignalSensor(TAG).getConsensuses();
    TigrScan tigrScan2("oasis","1.0",ef,transcriptId);
    queryGraph=&tigrScan2.parse(queryFile,queryIso,querySeq,querySeqStr,queryGC);
    numRightTermini=tigrScan1.numRightTermini();
    referenceGraph->setVertexIndices();
    queryGraph->setVertexIndices();
    processConfigFile(tigrScan1);

    //-------------------------------------------------------------------
    // Load the approximate alignments
    //-------------------------------------------------------------------
    cerr<<"loading approximate alignment from file "<<nucAppalFile<<"..."<<endl;
    nucAlignment=
      new ApproxNucAlignment<FLOAT_OR_DOUBLE>(nucAppalFile,*nucMatrix,
					      *referenceSeq,
					      referenceSeqStr->c_str(),*querySeq,
					      querySeqStr->c_str());

    //-------------------------------------------------------------------
    // Collect array of edges to use for random sampling
    //-------------------------------------------------------------------
    TigrVector<Edge*> xEdges, yEdges;
    collectEdges(*referenceGraph,xEdges);
    collectEdges(*queryGraph,yEdges);

    //-------------------------------------------------------------------
    // Pair off edges randomly and compute alignment scores
    //-------------------------------------------------------------------
    int nx=xEdges.size(), ny=yEdges.size();
    for(int i=0 ; i<numPoints ; ++i)
      {
	int ix=RandomNumber(nx), iy=RandomNumber(ny);
	Edge &ex=*xEdges[ix], &ey=*yEdges[iy];
	int xBegin=ex.getFeatureBegin(), xEnd=ex.getFeatureEnd();
	int yBegin=ey.getFeatureBegin(), yEnd=ey.getFeatureEnd();
	int xLen=xEnd-xBegin, yLen=yEnd-yBegin;
	float lenRatio=xLen/yLen;
	if(lenRatio<0.95 || lenRatio>1.053) {--i;continue;}
	//cout<<"================================================="<<endl;
	float approxScore=approxAlign(xBegin,xEnd,yBegin,yEnd,*nucAlignment);
	if(approxScore<.5) continue;
	float needlemanScore=needleAlign(xBegin,xEnd,yBegin,yEnd,
					 *referenceSeq,*querySeq,*nucMatrix,
					 gapPenalty);
	cout<<"SCORES: "<<approxScore<<"\t"<<needlemanScore<<endl;
	//if(needlemanScore<approxScore+0.1 && approxScore>=0.63) cout<<"BUZZARD ("<<xBegin<<"-"<<xEnd<<") = ("<<yBegin<<"-"<<yEnd<<")"<<endl;
      }

    return 0;
  }



/**************************************************************************
  This method computes the inductive score along a single genome's graph
  (left to right) and stores progressive values at signals along the path.
  This is necessary only because the code that produces GFF output needs
  to be able to compute the exon scores (log likelihood ratio) and it does
  this by subtracting one inductive score from the previous to get an exon
  coding score.
 */
void Oasis::computePathScore(TigrStack<Signal*> &path)
{
  // The top of the stack is at the beginning of the substrate, so we
  // need to iterate backward (=downward) through the stack:
  TigrStack<Signal*>::reverse_iterator cur=path.rbegin(), end=path.rend();
  Signal *pred=*cur;
  ++cur;
  for(; cur!=end ; ++cur)
    {
      Signal *signal=*cur;
      Edge *edge=signal->findEdgeInFrom(pred);
      if(!edge) throw "edge not found in Oasis::computePathScore()";
      int phase=getPhase(*signal);
      int predPhase=edge->propagateBackward(phase);
      double edgeInductiveScore=
	edge->isPhased() ?
	static_cast<PhasedEdge*>(edge)->getInductiveScore(predPhase) :
	static_cast<NonPhasedEdge*>(edge)->getInductiveScore();
      signal->getInductiveScore(phase)=
	edgeInductiveScore+signal->contextWindowScore();
      pred=signal;
    }
}



/**************************************************************************
 This method parses the promer file and allocates the cells of the 
 sparse dynamic programming matrix so as to explicitly represent only
 the parts of the matrix that make sense according to promer
 */
void Oasis::allocateMatrix(const TigrString &deltaFile,
				 int signalTolerance, 
				 int matchEdgeTolerance,int minIdent,
				 int minSim,
				 int maxStop,ParseGraph &referenceGraph,
				 ParseGraph &queryGraph,
				 float offMatchPortion)
{
  //--------------------------------------------------------------------
  // First, parse the promer matches out of the promer file 
  //--------------------------------------------------------------------
  const int numAminoAlignments=aminoAlignments.size();
  matrix=new OasisMatrix(referenceGraph,queryGraph);
  PromerParser promerParser(matchEdgeTolerance,minIdent,minSim,maxStop);
  PromerCoordMap *coordMap=promerParser.parseDelta(deltaFile);
  PromerSignalMap signalMap(referenceGraph,*coordMap,signalTolerance,
			    nonexactSignalTolerance,reductionLevel>=2);
  const int numReferenceSignals=referenceGraph.numVertices();
  const int numQuerySignals=queryGraph.numVertices();
  const int qLeftTermini=referenceGraph.countLeftTermini();
  const int rLeftTermini=queryGraph.countLeftTermini();
  const int qRightTermini=referenceGraph.countRightTermini();
  const int rRightTermini=queryGraph.countRightTermini();
  const int numColumns=numQuerySignals, numRows=numReferenceSignals;
  const int qBegin=qLeftTermini, rBegin=rLeftTermini;
  const int qEnd=numReferenceSignals-qRightTermini;
  const int rEnd=numQuerySignals-rRightTermini;
  if(coordMap->numMatches()==0) throw "No PROmer matches!";
  TigrArray1D<int> minRows(numColumns);
  TigrArray1D<int> maxRows(numColumns);
  minRows.setAllTo(0); maxRows.setAllTo(numRows-1);
  int numCellsAllocated=0;

  //--------------------------------------------------------------------
  // Allocate cells for the left & right termini
  //--------------------------------------------------------------------
  for(int i=0 ; i<rLeftTermini ; ++i)
    {
      if(referenceGraph.getIthVertex(i)->getSignalType()!=
	 queryGraph.getIthVertex(i)->getSignalType())
	{
	  cout<<i<<"\t"<<*referenceGraph.getIthVertex(i)<<" vs "
	      <<*queryGraph.getIthVertex(i)<<endl;
	  throw "Hmmmmm.... was expecting same signal types for left termini";
	}
      minRows[i]=maxRows[i]=i;
    }
  int i=numQuerySignals-1, j=numReferenceSignals-1;
  for(int k=0 ; k<rRightTermini ; ++k)
    {
      if(referenceGraph.getIthVertex(j)->getSignalType()!=
	 queryGraph.getIthVertex(i)->getSignalType())
	throw "Hmmmmm.... was expecting same signal types for right termini";
      minRows[i]=maxRows[i]=j;
      --i; --j;
    }

  if(reductionLevel>=1)
    {
      //--------------------------------------------------------------------
      // Now for any query signal which can be unamibiguously mapped by
      // one or more promer matches to a set of like reference signals, constrain
      // that signal's column to just the cells that pair it off with those
      // reference signals.  Also, if the mapping is not unamibiguous, select a
      // set of rows within +/- some number of base pairs from the mapped point.
      //--------------------------------------------------------------------
      for(int i=rBegin ; i<rEnd ; ++i)
	{
	  SignalPtr querySignal=queryGraph.getIthVertex(i);
	  TigrSet<Signal*> referenceSignals;
	  bool withinMatch=signalMap.mapSignal(*querySignal,referenceSignals);
	  int numHits=referenceSignals.size();
	  if(!withinMatch) continue;
	  Column &column=matrix->allocateColumn(i,numHits);
	  if(numHits==0) continue;
	  TigrSet<Signal*>::iterator cur=referenceSignals.begin(), 
	    end=referenceSignals.end();
	  for(int j=0; cur!=end ; ++cur, ++j)
	    {
	      SignalPtr referenceSignal=*cur;
	      column[j]=new Cell(*referenceSignal,*querySignal,numAminoAlignments);
	      ++numCellsAllocated;
	    }
	  column.sort();
	}
      
      if(reductionLevel>=3)
	{
	  //--------------------------------------------------------------------
	  // Now handle the signals that weren't mappable via promer.  Try to
	  // constrain their columns by observing whether their left and right
	  // neighbors in the graph are themselves constrained.
	  //--------------------------------------------------------------------  
	  for(int i=rBegin ; i<rEnd ; ++i)   // left to right, infer min #rows
	    {
	      Column *column=matrix->getColumn(i);
	      minRows[i]=column?
		(column->size() ? (*column)[0]->getReferenceIndex() : numRows) :
		getMinRowOnLeft(queryGraph.getIthVertex(i)->getEdgesIn(),
				minRows,numRows);
	    }	
	  for(int i=rEnd-1 ; i>=rBegin ; --i)// right to left, infer max #rows
	    {
	      Column *column=matrix->getColumn(i);
	      maxRows[i]=column?
		(column->size() ? (*column)[column->size()-1]->getReferenceIndex() 
		    :-1) :
		getMaxRowOnRight(queryGraph.getIthVertex(i)->getEdgesOut(),
				 maxRows,numRows);
	    }
	  
	  if(reductionLevel>=4)
	    {
	      //-------------------------------------------------------------------
	      // For signals not mapped via promer (and not constrained above by
	      // neighbors in the parse graph), find the nearest promer match and
	      // use the boundaries of the match (fuzzily) to establish constraints
	      // in that direction for these columns.  ("BLOCKING")
	      //-------------------------------------------------------------------
	      for(int i=rBegin ; i<rEnd ; ++i)
		if(!matrix->getColumn(i))
		  {
		    int &minRow=minRows[i], &maxRow=maxRows[i];
		    SignalPtr querySignal=queryGraph.getIthVertex(i);
		    PromerMatch *leftMatch, *rightMatch;
		    coordMap->getNearestMatches(querySignal->getConsensusPosition(),
						leftMatch,rightMatch);
		    if(/*minRow<1 &&*/ leftMatch)
		      {
			int matchLen=leftMatch->getQueryLength();
			int matchEnd=leftMatch->getQueryEnd();
			int margin=int(offMatchPortion*matchLen);
			int queryConstraint=matchEnd-margin;
			int referenceConstraint=leftMatch->map(queryConstraint);
			int newMinRow=signalMap.rowNearReferenceCoord(referenceConstraint);
			if(newMinRow>minRow) minRow=newMinRow; // AGGRESSIVE BLOCKING
		      }
		    if(/*maxRow>=numRows-1 &&*/ rightMatch)
		      {
			int matchLen=rightMatch->getQueryLength();
			int matchBegin=rightMatch->getQueryBegin();
			int margin=int(offMatchPortion*matchLen);
			int queryConstraint=matchBegin+margin;
			int referenceConstraint=rightMatch->map(queryConstraint);
			int newMaxRow=signalMap.rowNearReferenceCoord(referenceConstraint);
			if(newMaxRow<maxRow) maxRow=newMaxRow; // AGGRESSIVE BLOCKING
		      }
		  }
	    }
	}
    }
  
  //--------------------------------------------------------------
  // Now apply all the foregoing constraints to allocate the cells
  //--------------------------------------------------------------
  for(int i=0 ; i<numColumns ; ++i)
    if(!matrix->getColumn(i))
      {
	int &minRow=minRows[i], &maxRow=maxRows[i];
	Signal &querySignal=*queryGraph.getIthVertex(i);
	SignalType signalType=querySignal.getSignalType();
	int numRows=maxRow-minRow+1;
	Strand strand=querySignal.getStrand();

	// Don't need to match signals with those of different signal types
	int sameStrandRows=0;
	for(int j=minRow ; j<=maxRow ; ++j)
	  {
	    Signal &referenceSignal=*referenceGraph.getIthVertex(j);
	    if(referenceSignal.getSignalType()==signalType)// ENFORCE SIGNAL MATCH
	      ++sameStrandRows;
	  }
	
	Column &column=matrix->allocateColumn(i,sameStrandRows);
	for(int i=0, j=0 ; i<numRows ; ++i)
	  {
	    Signal &referenceSignal=*referenceGraph.getIthVertex(minRow+i);
	    if(referenceSignal.getSignalType()!=signalType) continue;// AS ABOVE
	    column[j]=new Cell(referenceSignal,querySignal,numAminoAlignments);
	    ++numCellsAllocated;
	    ++j;
	  }
      }

  
  //#########    PRINT OUT THE SPARSE DP MATRIX     ###########################
  if(!matrixOutFile.isEmpty())
    {
      ofstream os(matrixOutFile.c_str());
      for(int i=0 ; i<numColumns ; ++i)
	{
	  int minRow=minRows[i], maxRow=maxRows[i];
	  Signal &querySignal=*queryGraph.getIthVertex(i);
	  int numRows=maxRow-minRow+1;
	  Column *col=matrix->getColumn(i);
	  if(numRows>=0 && col && col->size())
	    {
	      os<<i<<" : "<<numRows<<"\t\t("<<minRow<<":"
		<< (*col)[0]->referenceSignal.getConsensusPosition() 
		<<" - "<<maxRow<<":"
		<< (*col)[col->size()-1]->referenceSignal.getConsensusPosition() 
		<<")"<<"   POS="<<querySignal.getConsensusPosition()<<"\t";
	      int n=col->size();
	      for(int i=0 ; i<n ; ++i)
		os<<(*col)[i]->referenceSignal.getIndex()<<',';
	      os<<endl;
	    }
	  /*
	  else if(col && col->size())
	    {
	      throw 0;
	    os<<i<<" : "<<col->size()<<"\t\t("<<(*col)[0]->getReferenceIndex()
	      <<":"<<(*col)[0]->referenceSignal.getConsensusPosition() <<" - "
	      <<(*col)[0]->getReferenceIndex()
	      <<":"<<(*col)[0]->referenceSignal.getConsensusPosition() 
	      <<")   POS="<<querySignal.getConsensusPosition()<<endl;
	    }
	  */
	  else os<<i<<" : 0"<<endl;
	}
    }
  //###########################################################################

  int percentAllocated=
    int(0.5+100*numCellsAllocated/float(numReferenceSignals*numQuerySignals));
  cout<<numCellsAllocated<<"/"<<(numReferenceSignals*numQuerySignals)
      <<" cells allocated = "<<percentAllocated<<"%"<<endl;

  delete coordMap; 
}



/**************************************************************************
  This method considers all the left neighbors of a vertex, and for
  each such neighbor it looks at the smallest row in that neighbor's
  column.  The smallest such row among all the neighbors is returned.
*/
int Oasis::getMinRowOnLeft(TigrSet<Edge*> &edgesIn,
				 TigrArray1D<int> &minRows,int numRows)
{

  int minRow=numRows;
  TigrSet<Edge*>::iterator cur=edgesIn.begin(), end=edgesIn.end();
  for(; cur!=end ; ++cur)
    {
      Edge &edge=**cur;
      Signal &querySignal=*edge.getLeft();
      int queryIndex=querySignal.getIndex();
      int row=minRows[queryIndex];
      if(row<minRow) minRow=row;
    }
  return minRow==numRows ? 0 : minRow;
}



/**************************************************************************
 This method considers all the right neighbors of a vertex, and for
 each such neighbor it looks at the largest row in that neighbor's
 column.  The largest such row among all the neighbors is returned.
 */
int Oasis::getMaxRowOnRight(TigrSet<Edge*> &edgesOut,
				  TigrArray1D<int> &maxRows,int numRows)
{
  int maxRow=-1;
  TigrSet<Edge*>::iterator cur=edgesOut.begin(), end=edgesOut.end();
  for(; cur!=end ; ++cur)
    {
      Edge &edge=**cur;
      Signal &querySignal=*edge.getRight();
      int queryIndex=querySignal.getIndex();
      int row=maxRows[queryIndex];
      if(row>maxRow) maxRow=row;
    }
  return maxRow<0 ? numRows-1 : maxRow;
}



/**************************************************************************
 This methods loads some parameters from TigrScan's configuration file
 */
void Oasis::processConfigFile(TigrScan &tigrScan)
{
  TigrConfigFile *config=tigrScan.getConfigFile();
  offMatchPortion=config->getFloatOrDie("off-match-portion");
  signalTolerance=config->getIntOrDie("signal-tolerance");
  nonexactSignalTolerance=config->getIntOrDie("nonexact-signal-tolerance");
  matchEdgeTolerance=config->getIntOrDie("match-edge-tolerance");
  minIdent=config->getIntOrDie("min-promer-identity");
  minSim=config->getIntOrDie("min-promer-similarity");
  maxStop=config->getIntOrDie("max-promer-stop-codons");
  reductionLevel=config->getIntOrDie("reduction-level");
  nucMatrix=loadSubstMatrix(config->lookupOrDie("nuc-substitution-matrix"),false);
  aminoMatrix=loadSubstMatrix(config->lookupOrDie("amino-substitution-matrix"),
			      true);
  gcContent=config->getIntOrDie("genome-GC-content")/100.0;
  gapPenalty=config->getIntOrDie("needleman-gap-penalty");

  // Alignment score distributions
  double alpha=config->getFloatOrDie("noncoding-alignment-EVD-alpha");
  double beta=config->getFloatOrDie("noncoding-alignment-EVD-beta");
  ExtremeValueDistribution evd(alpha,beta);
  noncodingAlignScores=new DiscreteLogDistr<double>(100,evd);
  double mu=config->getFloatOrDie("coding-alignment-mu");
  double sigma=config->getFloatOrDie("coding-alignment-sigma");
  NormalDistribution N(mu,sigma);
  codingAlignScores=new DiscreteLogDistr<double>(100,N);

  // Alignment null models for score distributions
  float atContent=1-gcContent;
  float pMatch=0.5*(gcContent*gcContent+atContent*atContent);
  noncodingNullAlignScores=new PoissonDistribution(int(pMatch*100+0.5));
  int expectedMatches=int(100*aminoMatrix->percentageGreaterThanZero()+0.5);
  cerr<<"Using lamba="<<expectedMatches<<" for amino Poisson distribution"<<endl;
  codingNullAlignScores=new PoissonDistribution(expectedMatches);
}



/**************************************************************************
 The dynamic programming algorithm.  Fills out the OasisMatrix from the
 lower left corner to the upper right, updating inductive scores and linking
 back to optimal predecessor.
 */
void Oasis::evaluateMatrix()
{
  // Iterate through all the columns of the matrix
  TigrProgress progress;
  int numCellsProcessed=0;
  progress.start(matrix->getNumCells());
  int numColumns=matrix->numColumns(), numRows=matrix->numRows();
  for(int col=0 ; col<numColumns ; ++col)
    {
      Column &column=*matrix->getColumn(col);
      SignalPtr querySignal=queryGraph->getIthVertex(col);
      TigrSet<Edge*> &queryEdges=querySignal->getEdgesIn();

      // Iterate through all the rows in this column (since the matrix is sparse)
      int columnSize=column.size();
      for(int colIndex=0 ; colIndex<columnSize ; ++colIndex)
	{
	  // ********************* Process this cell ***********************
	  Cell &cell=*column[colIndex];
	  int row=cell.getReferenceIndex();
	  SignalPtr referenceSignal=referenceGraph->getIthVertex(row);
	  TigrSet<Edge*> &referenceEdges=referenceSignal->getEdgesIn();
	  TigrArray1D<CellLink> links(9);

	  // Form the cartesian product of these two edges sets = the set of
	  // all potential predecessor cells for this cell:
	  TigrSet<Edge*>::iterator queryCur=queryEdges.begin(), queryEnd=queryEdges.end();
	  for(; queryCur!=queryEnd ; ++queryCur)
	    {
	      Edge &queryEdge=**queryCur;
	      int queryLeftIndex=queryEdge.getLeft()->getIndex();
	      Column &leftColumn=*matrix->getColumn(queryLeftIndex);
	      TigrSet<Edge*>::iterator qCur=referenceEdges.begin(),
		qEnd=referenceEdges.end();
	      for(; qCur!=qEnd ; ++qCur)
		{
		  Edge &referenceEdge=**qCur;
		  int referenceLeftIndex=referenceEdge.getLeft()->getIndex();
		  Cell *pred=leftColumn.lookupReferenceIndex(referenceLeftIndex);
		  if(pred) 
		    evaluatePredecessor(cell,*pred,&queryEdge,&referenceEdge,links);
		}
#ifdef ALLOW_INSERTIONS
	      // Also consider horizontal links (=no reference edge)
	      Cell *pred=leftColumn.lookupReferenceIndex(row);
	      if(pred)
		evaluatePredecessor(cell,*pred,&queryEdge,NULL,links);
#endif
	    }
#ifdef ALLOW_INSERTIONS
	  // Also consider vertical links (=no query edge)
	  TigrSet<Edge*>::iterator qCur=referenceEdges.begin(),
	    qEnd=referenceEdges.end();
	  for(; qCur!=qEnd ; ++qCur)
	    {
	      Edge &referenceEdge=**qCur;
	      int referenceLeftIndex=referenceEdge.getLeft()->getIndex();
	      Cell *pred=column.lookupReferenceIndex(referenceLeftIndex);
	      if(pred) 
		evaluatePredecessor(cell,*pred,NULL,&referenceEdge,links);
	    }
#endif

	  // Not all predecessors were instantiated in the sparse matrix, so not
	  // all links are valid -- install the valid ones into the cell (we save
	  // space by not installing the inactive ones)
	  int activeLinks=countActiveLinks(links);
	  TigrArray1D<CellLink> &cellLinks=cell.allocateLinks(activeLinks);
	  installLinks(links,cellLinks);
	}
      numCellsProcessed+=columnSize;
      if(col%3==0)
	cerr<<progress.getProgress(numCellsProcessed)<<endl;
    }
}



/**************************************************************************
 This method copies links from the 'from' array into the 'to' array, maintaining
 order.  Not all 'from' links are active, so we just copy the active ones.
 Space is already allocated in 'to' for the right number of links.
 */
void Oasis::installLinks(TigrArray1D<CellLink> &from,TigrArray1D<CellLink> &to)
{
  int nextSlot=0;
  for(int i=0 ; i<9 ; ++i)
    {
      CellLink &link=from[i];
      if(link.isActive())
	to[nextSlot++]=link;
    }

  /*this debugging clause shows that the ordering of installed links is correct!
  //###DEBUGGING
  for(int i=1 ; i<nextSlot ; ++i)
    if(to[i].getLinkSignature()<=to[i-1].getLinkSignature()) throw "bad!";
  //###
  */
}



/**************************************************************************
 Just count the links in this array which are marked as active.
*/
int Oasis::countActiveLinks(TigrArray1D<CellLink> &links)
{
  int n=links.size(), count=0;
  for(int i=0 ; i<n ; ++i)
    {
      CellLink &link=links[i];
      if(link.isActive()) 
	++count;
    }
  return count;
}



/**************************************************************************
                     PREDECESSOR EVALUATION, PHASE I
 This method begins the evaluation of a prospective predecessor cell for a 
 given cell in the dynamic programming matrix.  The computation is split into
 multiple functions simply because there are multiple edge types in the
 graphs for both genomes and there would be a complex nested if-then-else 
 structure to navigate otherwise.
 */
void Oasis::evaluatePredecessor(Cell &cell,Cell &pred,Edge *queryEdge,
				Edge *referenceEdge,TigrArray1D<CellLink> &links)
{
  //---------------------------------------------------------------------
  // Iterate through all possible phases of the edge in the query
  // sequence, observing edge scores.  For each iteration, call a method
  // to do likewise for the reference sequence edges.  We will thereby 
  // consider all combinations of predecessors in the two graphs (or at
  // least those present in the sparse dynamic programming matrix).
  //---------------------------------------------------------------------
  FLOAT_OR_DOUBLE querySignalScore=cell.querySignal.contextWindowScore();
  if(queryEdge && queryEdge->isPhased())
    {
      PhasedEdge &queryPhasedEdge=static_cast<PhasedEdge&>(*queryEdge);
      for(int rPhase=0 ; rPhase<3 ; ++rPhase)
	{
	  // Get edge score (includes duration & transition scores, too)
	  FLOAT_OR_DOUBLE queryEdgeScore=queryPhasedEdge.getEdgeScore(rPhase);
	  if(!finite(queryEdgeScore)) continue; // does this edge have that phase?
	  evaluateReferencePred(cell,pred,links,queryEdgeScore+querySignalScore,
			    referenceEdge,queryEdge,rPhase);
	}
    }
  else
    {
      // Get edge score (includes duration & transition scores, too)
      FLOAT_OR_DOUBLE queryEdgeScore=
	queryEdge ? static_cast<NonPhasedEdge&>(*queryEdge).getEdgeScore() : 0;
      int phase=
	queryEdge ? (queryEdge->getLeft()->getStrand()==PLUS_STRAND ? 0 : 2) : 0;
      if(finite(queryEdgeScore)) // does this edge have that phase?
	evaluateReferencePred(cell,pred,links,queryEdgeScore+querySignalScore,
			  referenceEdge,queryEdge,phase);
    }
}



/**************************************************************************
                     PREDECESSOR EVALUATION, PHASE II
 This method continues the evaluation of a putative predecessor cell by
 considering all possible phases of the edge in the reference sequence.  This
 method is in turn called once for each possible phase in the query
 sequence edge.
 */
void Oasis::evaluateReferencePred(Cell &cell,Cell &pred,TigrArray1D<CellLink> &links,
			      FLOAT_OR_DOUBLE queryEdgeAndSignalScore,
			      Edge *referenceEdge,
			      Edge *queryEdge,int oldQueryPhase)
{
  int newQueryPhase=
    queryEdge ? queryEdge->propagateForward(oldQueryPhase) : oldQueryPhase /* ### */;
  
  // Get the signal score in the reference genome
  FLOAT_OR_DOUBLE referenceSignalScore=cell.referenceSignal.contextWindowScore();

  // Handle all phases of the reference edge
  if(referenceEdge && referenceEdge->isPhased())
    {
      PhasedEdge &referencePhasedEdge=static_cast<PhasedEdge&>(*referenceEdge);
      for(int qPhase=0 ; qPhase<3 ; ++qPhase)
	{
	  FLOAT_OR_DOUBLE referenceEdgeScore=referencePhasedEdge.getEdgeScore(qPhase);
	  if(!finite(referenceEdgeScore)) continue;
	  FLOAT_OR_DOUBLE edgeAndSignalScores=
	    /*queryEdgeAndSignalScore+*/ referenceEdgeScore+referenceSignalScore;
	  evaluateAlignment(cell,pred,edgeAndSignalScores,links,oldQueryPhase,
			    newQueryPhase,qPhase,queryEdge,referenceEdge);
	}
    }
  else
    {
      FLOAT_OR_DOUBLE referenceEdgeScore=
	referenceEdge ? static_cast<NonPhasedEdge&>(*referenceEdge).getEdgeScore() : 0;
      if(finite(referenceEdgeScore))
	{
	  FLOAT_OR_DOUBLE edgeAndSignalScores=
	    /*queryEdgeAndSignalScore+*/referenceEdgeScore+referenceSignalScore;
	  int phase=
	    referenceEdge ? 
	    (referenceEdge->getLeft()->getStrand()==PLUS_STRAND ? 0 : 2) : 0;
	  evaluateAlignment(cell,pred,edgeAndSignalScores,links,oldQueryPhase,
			    newQueryPhase,phase,queryEdge,referenceEdge);
	}
    }
}



/**************************************************************************
                     PREDECESSOR EVALUATION, PHASE III
 This method continues the evaluate of a prospective predecessor cell in
 the dynamic programming matrix by accounting for the alignment score of
 any amino acids accumulated up to this point for a given selection of
 query/reference sequence phase.
 */
void Oasis::evaluateAlignment(Cell &cell,Cell &pred,
			      FLOAT_OR_DOUBLE edgeAndSignalScores,
			      TigrArray1D<CellLink> &links,int oldQueryPhase,
			      int newQueryPhase,int oldReferencePhase,
			      Edge *queryEdge,Edge *referenceEdge)
{
  CellLink *link=pred.getLink(oldQueryPhase,oldReferencePhase);

  // #### THIS SHOWS THAT THE LINK-PHASE RELATION IS CORRECT!
  //if(link && (link->getReferencePhase()!=oldReferencePhase || 
  //   link->getQueryPhase()!=oldQueryPhase)) throw "bad!!!"; 

  if(!link && !pred.leftTermini()) return; // orphaned
  int newReferencePhase=
    referenceEdge ? referenceEdge->propagateForward(oldReferencePhase) : oldReferencePhase;

  //-----------------------------------------------------------------------
  // Assess alignment score
  //-----------------------------------------------------------------------
  double alignmentProb=NEGATIVE_INFINITY;
  if(queryEdge->isCoding())
    {
      double alignmentScore=
	getAminoAlignmentScore(pred,cell,*referenceEdge,*queryEdge);
      if(!isinf(alignmentScore))
	alignmentProb=getAminoAlignmentProb(alignmentScore);
    }
  else
    {
      double alignmentScore=
	nucAlignment->score(referenceEdge->getFeatureBegin(),
			    queryEdge->getFeatureBegin(),
			    referenceEdge->getFeatureEnd(),
			    queryEdge->getFeatureEnd());
      if(!isinf(alignmentScore))
	alignmentProb=getNucAlignmentProb(alignmentScore);
    }
  FLOAT_OR_DOUBLE newInductiveScore=
    alignmentProb+edgeAndSignalScores+(link ? link->getInductiveScore() : 0);
  if(queryEdge->isCoding() && !isinf(alignmentProb)) cout<<(queryEdge->isCoding()?"C ":"NC ")<<"ALIGN: "<<alignmentProb<<" CONTENT: "<<edgeAndSignalScores<<" INDUCT: "<<(link?link->getInductiveScore():0)<<endl;

  //-----------------------------------------------------------------------
  // Finally, if the inductive score for this predecessor cell in the given
  // pair of phases is superior to the previously known optimum in those
  // phases, then install a link to this predecessor into the current cell
  // (ie., overwrite the previous optimal link with a new link)
  //-----------------------------------------------------------------------
  int linkIndex=CellLink::getLinkIndex(newReferencePhase,newQueryPhase);
  CellLink &newLink=links[linkIndex];
  if(!newLink.isActive() || newLink.getInductiveScore()<newInductiveScore)
    {
      newLink.setQueryIndex(pred.getQueryIndex());
      newLink.setReferenceIndex(pred.getReferenceIndex());
      newLink.setQueryPhase(newQueryPhase);
      newLink.setReferencePhase(newReferencePhase);
      newLink.setInductiveScore(newInductiveScore);
      newLink.setActive();
    }
}



/**************************************************************************
 Load a substitution matrix from file.
 */
SubstitutionMatrix<FLOAT_OR_DOUBLE> *Oasis::loadSubstMatrix(const TigrString &filename,
							    bool amino)
{
  return
    amino ?
    new SubstitutionMatrix<FLOAT_OR_DOUBLE>(filename,AminoAlphabet::global) :
    new SubstitutionMatrix<FLOAT_OR_DOUBLE>(filename,DnaAlphabet::global);
}



/**************************************************************************
 This method traces back from the upper-right-hand corner to the lower-left-
 hand corner of the dynamic programming matrix (ie., through the "2D trellis")
 to find the optimal path through both genomes simultaneously.
 */
void Oasis::traceBack(TigrStack<Signal*> &qPath,TigrStack<Signal*> &rPath)
{
  // Start at the cells in the upper-right corner (including any cell
  // both of whose signals are right termini)
  int width=matrix->numQuerySignals(), height=matrix->numReferenceSignals();
  int queryRightTerminiBegin=width-numRightTermini;
  int referenceRightTerminiBegin=height-numRightTermini;
  Cell *bestCell=NULL;
  FLOAT_OR_DOUBLE bestScore;
  CellLink *bestLink;
  for(int x=queryRightTerminiBegin ; x<width ; ++x)
    for(int y=referenceRightTerminiBegin ; y<height ; ++y)
      {
	Cell *cell=matrix->getCell(x,y);
	if(!cell) continue;
	TigrArray1D<CellLink> &links=*cell->links;
	int numLinks=links.size();
	for(int i=0 ; i<numLinks ; ++i)
	  {
	    CellLink &link=links[i];
	    FLOAT_OR_DOUBLE score=link.getInductiveScore();
	    if(!bestCell || score>bestScore) 
	      {
		bestCell=cell;
		bestLink=&link;
		bestScore=score;
	      }
	  }
      }
  if(!bestCell) {throw "WARNING: no links from terminal cells";}
  Cell *cell=bestCell;
  CellLink *link=bestLink;
  int referencePhase=link->getReferencePhase();
  int queryPhase=link->getQueryPhase();
  int q=link->getReferenceIndex(), r=link->getQueryIndex();
  SignalPtr referenceSignal=referenceGraph->getIthVertex(q);
  SignalPtr querySignal=queryGraph->getIthVertex(r);
  qPath.push(referenceSignal);
  rPath.push(querySignal);

  // Trace back until a cell is reached that has two left termini as signals
  while(!cell->leftTermini())
    {
      q=link->getReferenceIndex();
      r=link->getQueryIndex();
      SignalPtr referencePred=referenceGraph->getIthVertex(q);
      SignalPtr queryPred=queryGraph->getIthVertex(r);
      cell=matrix->getCell(r,q);
      bool referenceNew=(referenceSignal!=referencePred);
      if(referenceNew) 
	{
	  for(int i=0 ; i<3 ; ++i) referenceSignal->setPredecessor(i,NULL);
	  referenceSignal->setPredecessor(referencePhase,referencePred);
	}
      referencePhase=traceBackPhase(referencePhase,*referenceSignal,*referencePred);
      bool queryNew=(querySignal!=queryPred);
      if(queryNew) 
	{
	  for(int i=0 ; i<3 ; ++i) querySignal->setPredecessor(i,NULL);
	  querySignal->setPredecessor(queryPhase,queryPred);
	}
      queryPhase=traceBackPhase(queryPhase,*querySignal,*queryPred);
      link=cell->getLink(queryPhase,referencePhase);
      referenceSignal=referencePred;
      querySignal=queryPred;
      if(referenceNew) qPath.push(referenceSignal);
      if(queryNew) rPath.push(querySignal);
    }
}



/**************************************************************************
 Given a signal and the phase of that signal, this method computes the
 phase of this signal's immediate predecessor.  ### I think this function
 can be replaced by a call to Edge::propagateBackward(phase).  Need to make 
 sure...
 */
int Oasis::traceBackPhase(int phase,Signal &signal,Signal &pred)
{
  switch(pred.getSignalType())
    {
    case NEG_TAG:
    case NEG_POLYA:
    case NEG_ATG:     
    case NEG_PROM:    
      phase=2;
      break;
    case PROM:        
    case ATG:
    case TAG:         
    case POLYA:
      phase=0;
      break;
    case GT:
    case NEG_AG:
      break;
    case AG:
      {
	int distance=
	  signal.getConsensusPosition() -
	  pred.posOfBaseFollowingConsensus();
	phase=posmod(phase-distance);
      }
      break;
    case NEG_GT:
      {
	int distance=
	  signal.getConsensusPosition() -
	  pred.posOfBaseFollowingConsensus();
	phase=posmod(phase+distance);
      }
      break;
    }
  return phase;
}



/**************************************************************************
 Given that the 2D trellis has already been built and trace-back has been
 performed to identify optimal predecessors in the 2D trellis, this function
 identifies the phase of a given signal by simply noting which predecessor
 pointer is non-NULL.  It is not a general-purpose phase-determining function!
 */
int Oasis::getPhase(Signal &signal)
{
  int phase=(signal.getStrand()==FORWARD_STRAND ? 0 : 2);
  for(int i=0 ; i<3 ; ++i)
    if(signal.getPredecessor(i))
      phase=i;
  return phase;
}



/**************************************************************************
 Generate the GFF for the prediction in a single genome.  Called once for 
 each genome.
 */
void Oasis::output(TigrStack<Signal*> &path,const TigrString &filename,
		   TigrScan &tigrScan,const TigrString &species)
{
  // Invariant: the stack always contains at least 2 signals, and both
  // the top signal and the bottom signal are bogus (left & right termini)
  SignalPtr thisSignal=path.pop();
  SignalPtr firstSignal=thisSignal;
  TigrString substrateId=tigrScan.getSubstrateId();
  substrateId=substrateId.substitute(";","");
  
  cout.precision(4);
  cout << "##gff-version 2" << endl;
  cout << "##source-version " << PROGRAM_NAME << " " << VERSION << endl;
  cout << "##date " << getDateAndTime();
  cout << "##Type DNA" << endl;

  int transgrp=0;
  while(!path.isEmpty())
    {
      int phase=getPhase(*thisSignal);
      SignalType thisType=thisSignal->getSignalType();
      SignalPtr nextSignal=path.pop();
      SignalType nextType=nextSignal->getSignalType();
      int thisPos=thisSignal->getConsensusPosition();
      int nextPos=nextSignal->getConsensusPosition();
      if(thisPos<0) thisPos=0;
      double exonScore;

      // Compute exon score
      if(::beginsCoding(thisType)) 
	{
	  int begin=thisSignal->posOfBaseFollowingConsensus();
	  int end=nextSignal->posOfBaseFollowingConsensus();
	  int length=nextPos-begin;
	  int nextPhase=getPhase(*nextSignal);
	  ContentType contentType=
	    SignalTypeProperties::global.getContentType(thisType,nextType);
	  double lengthPenalty=
	    tigrScan.getDistribution(contentType).getLogP(length);
	  double numerator=
	    nextSignal->getInductiveScore(nextPhase)-
	    nextSignal->contextWindowScore() -
	    thisSignal->getInductiveScore(phase) -
	    tigrScan.getTransitionProbs()->getLogP(thisType,nextType) -
	    lengthPenalty;
	  double denominator=
	    tigrScan.getIntergenicSums()[begin+length-1]-
	    (begin ? tigrScan.getIntergenicSums()[begin-1] : 0);
	  exonScore=numerator-denominator;
	}

      switch(thisType)
	{
	case PROM:
	  if(thisSignal!=firstSignal)
	    cout << substrateId << '\t' << species 
		 << "\tpromoter\t"
		 << thisPos+1 << '\t' 
		 << min(thisPos+1+thisSignal->getConsensusLength(),
			tigrScan.getSeqLen())
		 << "\t." << "\t+\t.\t"
		 << "transgrp=" << transgrp << ";" << endl;
	  break;
	case NEG_PROM:
	  if(thisSignal!=firstSignal)
	    cout << substrateId << '\t' 
		 << species << "\tpromoter\t"
		 << thisPos+1 << '\t'
		 << min(thisPos+1+thisSignal->getConsensusLength(),
			tigrScan.getSeqLen())
		 << "\t." << "\t-\t" << phase << "\t"
		 << "transgrp=" << transgrp << ";" << endl;
	  break;
	case ATG: // initial-exon or single-exon
	  ++transgrp;
	  switch(nextType)
	    {
	    case GT:
	      cout << substrateId << '\t' 
		   << species << "\tinitial-exon\t"
		   << thisPos+1 << '\t' 
		   << min(nextPos,tigrScan.getSeqLen()) << "\t" 
		   << exonScore << "\t+\t"
		   << phase << "\t"
		   << "transgrp=" << transgrp << ";" << endl;
	      break;
	    case TAG:
	      cout << substrateId << '\t' 
		   << species << "\tsingle-exon\t"
		   << thisPos+1 << '\t' 
		   << min(nextPos+3,tigrScan.getSeqLen()) << "\t" 
		   << exonScore << "\t+\t"
		   << phase << "\t"
		   << "transgrp=" << transgrp << ";" << endl;
	      break;
	    }
	  break;
	case AG: // internal-exon or final-exon
	  switch(nextType)
	    {
	    case GT:
	      cout << substrateId << '\t' << species 
		   << "\tinternal-exon\t"
		   << thisPos+3 << '\t' 
		   << min(nextPos,tigrScan.getSeqLen()) << "\t" 
		   << exonScore << "\t+\t"
		   << phase << "\t"
		   << "transgrp=" << transgrp << ";" << endl;
	      break;
	    case TAG:
	      cout << substrateId << '\t' 
		   << species << "\tfinal-exon\t"
		   << thisPos+3 << '\t' 
		   << min(nextPos+3,tigrScan.getSeqLen()) << "\t" 
		   << exonScore << "\t+\t"
		   << phase << "\t"
		   << "transgrp=" << transgrp << ";" << endl;
	      break;
	    }
	  break;
	case NEG_TAG: // final-exon or single-exon
	  ++transgrp;
	  switch(nextType)
	    {
	    case NEG_AG:
	      cout << substrateId << '\t' 
		   << species << "\tfinal-exon\t"
		   << thisPos+1 << '\t' 
		   << min(nextPos,tigrScan.getSeqLen()) << "\t" 
		   << exonScore << "\t-\t"
		   << phase << "\t"
		   << "transgrp=" << transgrp << ";" << endl;
	      break;
	    case NEG_ATG:
	      cout << substrateId << '\t' 
		   << species << "\tsingle-exon\t"
		   << thisPos+1 << '\t' 
		   << min(nextPos+3,tigrScan.getSeqLen()) 
		   << "\t" << exonScore << "\t-\t"
		   << phase << "\t"
		   << "transgrp=" << transgrp << ";" << endl;
	      break;
	    }
	  break;
	case NEG_GT: // internal-exon or initial-exon
	  switch(nextType)
	    {
	    case NEG_AG:
	      cout << substrateId << '\t' << species 
		   << "\tinternal-exon\t"
		   << thisPos+3 << '\t' 
		   << min(nextPos,tigrScan.getSeqLen()) 
		   << "\t" << exonScore << "\t-\t"
		   << phase << "\t"
		   << "transgrp=" << transgrp << ";" << endl;
	      break;
	    case NEG_ATG:
	      cout << substrateId << '\t' 
		   << species << "\tinitial-exon\t"
		   << thisPos+3 << '\t' 
		   << min(nextPos+3,tigrScan.getSeqLen()) 
		   << "\t" << exonScore << "\t-\t"
		   << phase << "\t"
		   << "transgrp=" << transgrp << ";" << endl;
	      break;
	    }
	  break;
	}
      thisSignal=nextSignal;
    }

  cout << tigrScan.getGffIsochoreMsg() << endl;
  cout << "# sequence length: " << tigrScan.getSeqLen() << endl;
  cout << "# For information about this file format, see:" << endl;
  cout << "#   <http://www.sanger.ac.uk/Software/formats/GFF/GFF_Spec.shtml>" 
     << endl;
}



/**************************************************************************
 Debugging.  Given a set of transcripts loaded from each of two GFF files
 (one per genome), we try to find all the signals for those transcripts
 in the parse graph and report which were found and what were their scores.
 */
void Oasis::debugGFF(TigrVector<TigrGffTranscript*> &qTranscripts,
		     TigrVector<TigrGffTranscript*> &rTranscripts,
		     ParseGraph &referenceGraph,ParseGraph &queryGraph,
		     TigrScan &qTigrScan,TigrScan &rTigrScan,
		     Sequence &qSeq,const TigrString &qSeqStr,
		     Sequence &rSeq,const TigrString &rSeqStr)
{
  GffPathFromParseGraph qPathfinder(qTigrScan), rPathfinder(rTigrScan);
  TigrVector<BOOL> qFound, rFound;
  TigrVector<SignalPtr> &qPath=
    *qPathfinder.getPathFromGff(qTranscripts,qSeq,qSeqStr,qFound);
  TigrVector<SignalPtr> &rPath=
    *rPathfinder.getPathFromGff(rTranscripts,rSeq,rSeqStr,rFound);
  debugGFF(rPath,rFound);
  debugGFF(qPath,qFound);
  if(rPath.size()!=qPath.size())
    cout<<"Can't probe for cells, due to inserted introns"<<endl;
  else
    {
      int n=rPath.size();
      for(int i=0 ; i<n ; ++i)
	{
	  SignalPtr rSignal=rPath[i], qSignal=qPath[i];
	  int rIndex=rSignal->getIndex();
	  int qIndex=qSignal->getIndex();
	  bool found=(bool) matrix->getCell(rIndex,qIndex);
	  if(!found)
	    cout<<"***MISSING CELL FOR "
		<<rSignal->getSignalType()
		<<"@"<<rSignal->getConsensusPosition()<<" x "
		<<qSignal->getSignalType()
		<<"@"<<qSignal->getConsensusPosition()
		<<endl;
	  else
	    cout<<"found cell for "
		<<rSignal->getSignalType()
		<<"@"<<rSignal->getConsensusPosition()<<" x "
		<<qSignal->getSignalType()
		<<"@"<<qSignal->getConsensusPosition()
		<<endl;
	}
    }
}


/**************************************************************************
 Continuation of above function.
 */
void Oasis::debugGFF(TigrVector<SignalPtr> &path,TigrVector<BOOL> &found)
{
  int n=path.size();
  for(int i=0 ; i<n ; ++i)
    {
      SignalPtr signal=path[i];
      double score=signal->contextWindowScore();
      int pos=signal->getConsensusPosition();
      SignalType signalType=signal->getSignalType();
      TigrString f=(found[i] ? "found" : "not found");
      cout<<signal->getIndex()<<"("<<pos<<")\t"
	  <<signalType<<"\t"<<score<<"\t"<<f<<endl;
    }
}



/**************************************************************************
 Loads the ApproxAminoAlignment objects from the promer/appal file.
 */
void Oasis::loadApproxAminoAligns(const TigrString &filename,
				  const SubstitutionMatrix<FLOAT_OR_DOUBLE> &M,
				  const Sequence &seq1,const char *seqStr1,
				  const Sequence &seq2,const char *seqStr2)
{
  ifstream is(filename.c_str());
  if(!is.good()) throw TigrString("Error loading file: ")+filename;
  while(!is.eof())
    {
      ApproximateAlignment<FLOAT_OR_DOUBLE> *a=
	ApproxAminoAlignment<FLOAT_OR_DOUBLE>::load(is,M,seq1,seqStr1,seq2,
						    seqStr2,*stopCodons);
      if(!a) break;
      aminoAlignments.push_back(a);
    }
  const int numAminoAlignments=aminoAlignments.size();
  tempAlignmentSet.setSize(numAminoAlignments);
}



/**************************************************************************
 We register each ApproxAminoAlignment with each of the OASIS cells that it
 covers, so that later when we have two cells and wish to find the alignments
 (promer hits) that connect the two we have only to take an intersection
 between their registered alignments, and this is very fast since we are
 using bitmaps to represent those sets (the intersection is done with a 
 bitwise AND).
 */
void Oasis::registerAminoAlignments()
{
  TigrVector<ApproximateAlignment<FLOAT_OR_DOUBLE>*>::iterator 
    cur=aminoAlignments.begin(), end=aminoAlignments.end();
  for(int i=0; cur!=end ; ++cur, ++i)
    {
      ApproximateAlignment<FLOAT_OR_DOUBLE> *alignment=*cur;
      int beginX=alignment->getBeginX(), endX=alignment->getEndX();
      int beginY=alignment->getBeginY(), endY=alignment->getEndY();
      registerAminoAlignment(i,beginY,endY,beginX,endX,*queryGraph); // ### x? y?
    }
}



/**************************************************************************
 Continuation of registerAminoAlignments().  Registers given alignment
 with all cells in the matrix either falling in [beginX,endX] on the
 one axis or [beginY,endY] on the other axis.
 */
void Oasis::registerAminoAlignment(int alignmentId,int beginX,int endX,
				   int beginY,int endY,ParseGraph &graph)
{
  // Find the leftmost signal in this graph which might fall under this
  // alignment
  int first=graph.findFirstSignalAfter(beginX), last=graph.numVertices();

  // For each column in [beginX,endX], register this alignment with all 
  // the cells in that column
  int i;
  for(i=first ; i<last ; ++i)
    {
      SignalPtr signal=graph.getIthVertex(i);
      int pos=signal->getConsensusPosition();
      if(pos>=endX) break;
      if(pos<beginX) throw "UH-OH!";//###

      // This signal falls under this alignment, so register the alignment
      // with all the OASIS cells in this signal's column of the matrix
      Column *column=matrix->getColumn(i);
      if(!column) continue;
      int n=column->size();
      for(int i=0 ; i<n ; ++i)
	(*column)[i]->promerHits.addMember(alignmentId);
    }

  // For each column NOT in [beginX,endX], register this alignment with
  // any cell in that column that falls within [beginY,endY]
  for(; i<last ; ++i)
    registerRows(*matrix->getColumn(i),alignmentId,beginY,endY);
  for(i=0 ; i<first ; ++i)
    registerRows(*matrix->getColumn(i),alignmentId,beginY,endY);
}



/**************************************************************************
 Continuation of registerAminoAlignment().  Registers the given alignment
 with all cells in a given column that fall within [begin,end].
 */
void Oasis::registerRows(Column &column,int alignmentId,
			 int beginY,int endY)
{
  // Perform binary search for first cell after beginY
  int columnSize=column.size();
  int first=0, last=columnSize;
  while(first<last)
    {
      int mid=(first+last)/2;
      Cell *cell=column[mid];
      int cellPos=cell->referenceSignal.getConsensusPosition();//###ref/query?
      if(beginY>cellPos) first=mid+1;
      else last=mid;
    }

  // The binary search may leave us 1 cell too far to the left, so
  // check for that case
  if(first<columnSize && 
     beginY>column[first]->referenceSignal.getConsensusPosition())
    ++first;

  // Register with all cells up to endY
  for(int i=first ; i<columnSize ; ++i)
    {
      Cell *cell=column[i];
      if(cell->referenceSignal.getConsensusPosition()>endY) break;
      //if(cell->referenceSignal.getConsensusPosition()<beginY) throw "OK";
      cell->promerHits.addMember(alignmentId);
    }
}



/**************************************************************************
 */
double Oasis::getAminoAlignmentScore(Cell &predCell,Cell &thisCell,
				     Edge &refEdge,Edge &queryEdge)
{
  // Get all the promer alignments covering both cells
  predCell.promerHits.intersect(thisCell.promerHits,tempAlignmentSet);

  // Choose the best-scoring alignment
  int n=aminoAlignments.size();
  double bestScore=NEGATIVE_INFINITY;
  for(int i=0 ; i<n ; ++i)
    {
      if(tempAlignmentSet.isMember(i))
	{
	  ApproximateAlignment<FLOAT_OR_DOUBLE> *aln=aminoAlignments[i];



	  //### DON'T FORGET TO CHECK FOR MATCHING PHASE!!!!!

	  
	  
	  double score=aln->score(refEdge.getFeatureBegin(),
				  queryEdge.getFeatureBegin(),
				  refEdge.getFeatureEnd(),
				  queryEdge.getFeatureEnd());
	  if(score>bestScore) bestScore=score;
	}
    }
  return bestScore;
}



/**************************************************************************
 We compute P(orth|alignment score) using Bayes' Rule:
       P(orth|score)=P(score|orth)*P(orth)/P(score)
 We omit P(orth) from the evaluation since it is unchanging over the argmax.
 P(score|orth) is modeled from training data using a normal distribution,
 and P(score) is modeled using a Poisson distribution assuming spurious match 
 probability based on GC content of entire genome.
 */
double Oasis::getAminoAlignmentProb(double alignmentScore)
{
  int scoreX100=int(100*alignmentScore+0.5);
  double logP_score_given_orth=(*codingAlignScores)[scoreX100]
    -log(codingAlignScores->getProbOfMean());
  //double logP_score=codingNullAlignScores->logDensity(int(scoreX100+0.5));

  cout<<"logP[score|orth]="<<(*codingAlignScores)[scoreX100]<<", mean="<<codingAlignScores->getMean()<<", logP[mean]="<<log(codingAlignScores->getProbOfMean())<<endl;
  return logP_score_given_orth;//-logP_score;
}



/**************************************************************************
 We compute P(orth|alignment score) using Bayes' Rule:
       P(orth|score)=P(score|orth)*P(orth)/P(score)
 We omit P(orth) from the evaluation since it is unchanging over the argmax.
 P(score|orth) is modeled from training data using an Extreme Value 
 Distribution, and P(score) is modeled using a Poisson distribution with the
 match probability computed from the genome's GC content.
 */
double Oasis::getNucAlignmentProb(double alignmentScore)
{
  int scoreX100=int(100*alignmentScore+0.5);
  double logP_score_given_orth=(*noncodingAlignScores)[scoreX100]
    -log(noncodingAlignScores->getProbOfMean());
  //double logP_score=noncodingNullAlignScores->logDensity(int(scoreX100+0.5));

  return logP_score_given_orth;//-logP_score;
}



/**************************************************************************
 */
void collectEdges(ParseGraph &G,TigrVector<Edge*> &E)
{
  int n=G.numVertices();
 NEXT:
  for(int i=0 ; i<n ; ++i)
    {
      SignalPtr signal=G.getIthVertex(i);
      TigrSet<Edge*> &edges=signal->getEdgesOut();
      TigrSet<Edge*>::iterator cur=edges.begin(), end=edges.end();
      for(; cur!=end ; ++cur)
	{
	  Edge *edge=*cur;
	  if(!edge->isCoding()) break;
	  E.push_back(edge);
	}
    }
}



double approxAlign(int xBegin,int xEnd,int yBegin,int yEnd,
		   ApproximateAlignment<FLOAT_OR_DOUBLE> &A)
{
  return A.score(xBegin,yBegin,xEnd,yEnd);
}



double needleAlign(int xBegin,int xEnd,int yBegin,int yEnd,
		   Sequence &s1,Sequence &s2,
		   const SubstitutionMatrix<FLOAT_OR_DOUBLE> &M,
		   FLOAT_OR_DOUBLE gapExtendPenalty)
{
  int xLen=xEnd-xBegin, yLen=yEnd-yBegin;
  //cerr<<"Running Needleman..."<<xLen<<"x"<<yLen<<endl;
  Sequence sub1, sub2;
  s1.getSubsequence(xBegin,xLen,sub1);
  s2.getSubsequence(yBegin,yLen,sub2);
  Needleman<FLOAT_OR_DOUBLE> aligner(DnaAlphabet::global,sub1,sub2,
				     M,-2/*gapExtendPenalty*/);
  Alignment &alignment=*aligner.fullAlignment();
  int len=alignment.getAlignmentLength();
  int mismatches, insertions;
  alignment.countMismatches(mismatches,insertions);
  double matches=len-insertions-mismatches;
  cout<<"   Needleman done.  "<<matches<<" matches / "<<len<<endl;
  cout<<" "<<int((sub1.countOccurrences(DnaAlphabet::global.lookup('A'))/float(xLen)*100+.5))
      <<"A "<<int((sub1.countOccurrences(DnaAlphabet::global.lookup('C'))/float(xLen)*100+.5))
      <<"C "<<int((sub1.countOccurrences(DnaAlphabet::global.lookup('G'))/float(xLen)*100+.5))
      <<"G "<<int((sub1.countOccurrences(DnaAlphabet::global.lookup('T'))/float(xLen)*100+.5))
      << "T"<<endl;
  cout<<"  "<<mismatches<<" mismatches, "<<insertions<<" insertions, "<<len<<" len, "<<matches<<" matches"<<endl;
  return matches/len;
}



